"""Utility functions."""

import datetime
import ifaddr
import socket
import asyncio

from contextlib import contextmanager
from threading import Thread


def timestamp():
    """
    Get the current time.
    Returns the current time in the form YYYY-mm-ddTHH:MM:SS+00:00
    """
    return datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S+00:00")


def get_ip():
    """
    Get the default local IP address.
    From: https://stackoverflow.com/a/28950776
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(("10.255.255.255", 1))
        ip = s.getsockname()[0]
    except (socket.error, IndexError):
        ip = "127.0.0.1"
    finally:
        s.close()

    return ip


def get_addresses():
    """
    Get all IP addresses.
    Returns list of addresses.
    """
    addresses = set()

    for iface in ifaddr.get_adapters():
        for addr in iface.ips:
            # Filter out link-local addresses.
            if addr.is_IPv4:
                ip = addr.ip

                if not ip.startswith("169.254."):
                    addresses.add(ip)
            elif addr.is_IPv6:
                # Sometimes, IPv6 addresses will have the interface name
                # appended, e.g. %eth0. Handle that.
                ip = addr.ip[0].split("%")[0].lower()

                if not ip.startswith("fe80:"):
                    addresses.add(f"[{ip}]")

    return sorted(list(addresses))


@contextmanager
def background_thread_loop():
    def run_forever(loop):
        asyncio.set_event_loop(loop)
        loop.run_forever()

    loop = asyncio.new_event_loop()
    try:
        thread = Thread(target=run_forever, args=(loop,))
        thread.start()
        yield loop
    finally:
        loop.call_soon_threadsafe(loop.stop)
        thread.join()



"""Python Web Thing server implementation."""

import socket

from zeroconf import ServiceInfo, Zeroconf

from starlette.routing import Route, WebSocketRoute, Mount
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.middleware.authentication import AuthenticationMiddleware

from .handlers import (
    ThingHandler,
    WsThingHandler,
    ThingsHandler,
    PropertyHandler,
    EventHandler,
    PropertiesHandler,
    ActionHandler,
    ActionsHandler,
    ActionIDHandler,
    EventsHandler,
)
from .containers import (
    MultipleThings, SingleThing, DevicePairingEvent, DeviceRemoveEvent
)
from .utils import get_addresses, get_ip
from .mixins import AsyncMixin
from .thing import Thing
from .value import Value
from .property import Property


class DefaultHeaderMiddleware(BaseHTTPMiddleware):
    """Set the default headers for all requests."""

    async def dispatch(self, request, call_next):
        response = await call_next(request)
        response.headers["Access-Control-Allow-Origin"] = "*"
        response.headers[
            "Access-Control-Allow-Headers"
        ] = "Origin, X-Requested-With, Content-Type, Accept"
        response.headers[
            "Access-Control-Allow-Methods"
        ] = "GET, HEAD, PUT, POST, DELETE"

        return response


middlewares = [
    Middleware(DefaultHeaderMiddleware),
    Middleware(TrustedHostMiddleware, allowed_hosts=["*"]),
]


class Server(Thing):
    type = ["Server"]
    description = "Webthing Server"

    def __init__(self):
        super().__init__(
            "urn:webthing:server",
            "Webthing Server",
        )

    async def build(self):
        await self.add_property(
            Property(
                "state",
                Value("ON"),
                metadata={
                    "@type": "ServerStateProperty",
                    "title": "State",
                    "type": "string",
                    "enum": ["ON", "OFF", "REBOOT"],
                    "description": "state of webthing server",
                },
            )
        )

        await self.add_available_event(
            DevicePairingEvent,
            {
                "description": "new device",
                "type": "object",
                "required": ["@type", "id", "title"],
                "properties": {
                    "@type": {
                        "type": "array",
                    },
                    "id": {
                        "type": "string",
                    },
                    "title": {
                        "type": "string",
                    },
                },
            }
        )

        await self.add_available_event(
            DeviceRemoveEvent,
            {
                "description": "device removed event",
                "type": "object",
                "required": ["id", "title"],
                "properties": {
                    "id": {
                        "type": "string",
                    },
                    "title": {
                        "type": "string",
                    },
                },
            }
        )

        return self


class WebThingServer(AsyncMixin):
    """Server to represent a Web Thing over HTTP."""

    def __init__(
            self,
            loop,
            thing_cls=None,
            port=8000,
            hostname=None,
            base_path="",
            additional_routes=None,
            additional_middlewares=None,
            additional_on_startup=None,
            additional_on_shutdown=None,
    ):
        """
        Initialize the WebThingServer.
        For documentation on the additional route format, see:
        https://www.starlette.io/applications/
        loop -- event loop
        thing_cls -- make things managed by this server -- should be of type
                  Thing or List of Things
        port -- port to listen on (defaults to 8000)
        hostname -- Optional host name, i.e. mything.com
        base_path -- base URL path to use, rather than '/'
        additional_routes -- list of additional routes to add to the server
        additional_middlewares -- list of additional middlewares
        additional_on_startup -- list of additional starup event handlers
        additional_on_shutdown -- list of additional shutdown event handlers
        """
        self._loop = loop
        server = self._run_async(Server().build())
        self.things = MultipleThings({server.id: server}, "things")
        self.thing_cls = thing_cls if isinstance(thing_cls, list) else [thing_cls]
        self.port = port
        self.hostname = hostname
        self.base_path = base_path.rstrip("/") if base_path else "/things"
        self.additional_routes = additional_routes
        self.additional_middlewares = additional_middlewares
        self.additional_on_startup = [] if additional_on_startup is None else list(additional_on_startup)
        self.additional_on_shutdown = [] if additional_on_shutdown is None else list(additional_on_shutdown)
        system_hostname = socket.gethostname().lower()

        self.hosts = [
            "localhost",
            f"localhost:{self.port}",
            f"{system_hostname}.local",
            f"{system_hostname}.local:{self.port}",
        ]

        for address in get_addresses():
            self.hosts.extend(
                [address, f"{address}:{self.port}", ]
            )

        if self.hostname is not None:
            self.hostname = self.hostname.lower()
            self.hosts.extend(
                [self.hostname, f"{self.hostname}:{self.port}", ]
            )

    def create(self):
        return self._run_async(self._create())

    async def config_things(self):
        for cls in self.thing_cls:
            thing = await cls().build()
            await self.things.add_thing(thing)

    async def config_routes(self):
        if isinstance(self.things, MultipleThings):
            for idx, thing in await self.things.get_things():
                await thing.set_href_prefix(f"{self.base_path}/{idx}")

            base_route = [
                Route(f"{self.base_path}",
                      ThingsHandler),
            ]
            routes = [
                Route("/{thing_id:str}",
                      ThingHandler),
                WebSocketRoute("/{thing_id:str}",
                               WsThingHandler),
                Route("/{thing_id:str}/properties",
                      PropertiesHandler),
                Route("/{thing_id:str}/properties/{property_name:str}",
                      PropertyHandler),
                Route("/{thing_id:str}/actions",
                      ActionsHandler),
                Route("/{thing_id:str}/actions/{action_name:str}",
                      ActionHandler),
                Route("/{thing_id:str}/actions/{action_name:str}/{action_id}",
                      ActionHandler),
                Route("/{thing_id:str}/events",
                      EventHandler),
                Route("/{thing_id:str}/events/{event_name:str}",
                      EventHandler),
            ]
        else:
            thing = await self.things.get_thing()
            await thing.set_href_prefix(self.base_path)

            base_route = [
                Route(f"{self.base_path}",
                      ThingHandler),
                WebSocketRoute(f"{self.base_path}",
                               WsThingHandler),
            ]
            routes = [
                Route("/properties",
                      PropertiesHandler),
                Route("/properties/{property_name:str}",
                      PropertyHandler),
                Route("/actions",
                      ActionsHandler),
                Route("/actions/{action_name:str}",
                      ActionHandler),
                Route("/actions/{action_name:str}/{action_id:str}",
                      ActionIDHandler),
                Route("/events",
                      EventsHandler),
                Route("/events/{event_name:str}",
                      EventHandler),
            ]

        if isinstance(self.additional_routes, list):
            routes = self.additional_routes + routes

        if self.base_path:
            routes = base_route + [
                # Route('/', homepage),
                Mount(f"{self.base_path}", routes=routes),
            ]

        return routes

    async def config_middlewares(self):
        middlewares = []
        if self.additional_middlewares:
            assert isinstance(self.additional_middlewares, list)
            middlewares.extend(self.additional_middlewares)

        return middlewares

    async def config_on_startups(self):
        on_startups = [self.start]
        if self.additional_on_startup:
            on_startups.extend(self.additional_on_startup)

        return on_startups

    async def config_on_shutdowns(self):
        on_shutdowns = [self.stop]
        if self.additional_on_shutdown:
            assert isinstance(self.additional_on_shutdown, list)
            on_shutdowns.extend(self.additional_on_shutdown)

        return on_shutdowns

    async def _create(self):
        await self.config_things()
        routes = await self.config_routes()
        middlewares = await self.config_middlewares()
        on_startups = await self.config_on_startups()
        on_shutdowns = await self.config_on_shutdowns()

        app = Starlette(
            debug=True, routes=routes, middleware=middlewares, on_startup=on_startups, on_shutdown=on_shutdowns,
        )

        app.state.things = self.things

        require_auth = False
        for middleware in middlewares:
            if middleware.cls == AuthenticationMiddleware:
                require_auth = True

        app.state.require_auth = require_auth

        return app

    async def start(self):
        """Start listening for incoming connections."""
        name = await self.things.get_name()
        args = [
            '_webthing._tcp.local.',
            f"{name}._webthing._tcp.local.",
        ]
        kwargs = {'port': self.port, 'properties': {
            'path': '/',
        }, 'server': '{}.local.'.format(socket.gethostname()), 'addresses': [socket.inet_aton(get_ip())]}
        self.service_info = ServiceInfo(*args, **kwargs)
        self.zeroconf = Zeroconf()
        self.zeroconf.register_service(self.service_info)

    async def stop(self):
        """Stop listening."""
        self.zeroconf.unregister_service(self.service_info)
        self.zeroconf.close()



import jwt
import asyncio
import functools
import inspect
import typing

from starlette.exceptions import HTTPException
from starlette.requests import HTTPConnection, Request
from starlette.responses import RedirectResponse, Response
from starlette.authentication import (
    AuthenticationBackend, AuthenticationError, BaseUser, AuthCredentials,
    UnauthenticatedUser)
from starlette.websockets import WebSocket, WebSocketDisconnect


def has_required_scope(conn: HTTPConnection, scopes: typing.Sequence[str]) -> bool:
    for scope in scopes:
        if scope not in conn.auth.scopes:
            return False
    return True


def requires(
    scopes: typing.Union[str, typing.Sequence[str]],
    status_code: int = 403,
    redirect: str = None,
) -> typing.Callable:
    scopes_list = [scopes] if isinstance(scopes, str) else list(scopes)

    def decorator(func: typing.Callable) -> typing.Callable:
        type = None
        sig = inspect.signature(func)
        for idx, parameter in enumerate(sig.parameters.values()):
            if parameter.name == "request" or parameter.name == "websocket":
                type = parameter.name
                break
        else:
            raise Exception(
                f'No "request" or "websocket" argument on function "{func}"'
            )

        if type == "websocket":
            # Handle websocket functions. (Always async)
            @functools.wraps(func)
            async def websocket_wrapper(
                *args: typing.Any, **kwargs: typing.Any
            ) -> None:
                websocket = kwargs.get("websocket", args[idx])
                assert isinstance(websocket, WebSocket)

                if websocket.app.state.require_auth and not has_required_scope(websocket, scopes_list):
                    await websocket.close()
                    raise WebSocketDisconnect
                else:
                    await func(*args, **kwargs)

            return websocket_wrapper

        elif asyncio.iscoroutinefunction(func):
            # Handle async request/response functions.
            @functools.wraps(func)
            async def async_wrapper(
                *args: typing.Any, **kwargs: typing.Any
            ) -> Response:
                request = kwargs.get("request", args[idx])
                assert isinstance(request, Request)

                if request.app.state.require_auth and not has_required_scope(request, scopes_list):
                    if redirect is not None:
                        return RedirectResponse(
                            url=request.url_for(redirect), status_code=303
                        )
                    raise HTTPException(status_code=status_code)
                return await func(*args, **kwargs)

            return async_wrapper

        else:
            # Handle sync request/response functions.
            @functools.wraps(func)
            def sync_wrapper(*args: typing.Any, **kwargs: typing.Any) -> Response:
                request = kwargs.get("request", args[idx])
                assert isinstance(request, Request)

                if request.app.state.require_auth and not has_required_scope(request, scopes_list):
                    if redirect is not None:
                        return RedirectResponse(
                            url=request.url_for(redirect), status_code=303
                        )
                    raise HTTPException(status_code=status_code)
                return func(*args, **kwargs)

            return sync_wrapper

    return decorator


class JWTUser(BaseUser):
    def __init__(self, username: str, token: str, payload: dict) -> None:
        self.username = username
        self.token = token
        self.payload = payload

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.username


class JWTAuthenticationBackend(AuthenticationBackend):

    def __init__(self, secret_key: str, algorithm: str = 'HS256', prefix: str = 'JWT', username_field: str = 'username'):
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.prefix = prefix
        self.username_field = username_field

    @classmethod
    def get_token_from_header(cls, authorization: str, prefix: str):
        """
        Parses the Authorization header and returns only the token
        :param authorization:
        :return:
        """
        try:
            scheme, token = authorization.split()
        except ValueError:
            raise AuthenticationError('Could not separate Authorization scheme and token')
        if scheme.lower() != prefix.lower():
            raise AuthenticationError(f'Authorization scheme {scheme} is not supported')
        return token

    async def authenticate(self, request):
        if "Authorization" not in request.headers:
            return None
        auth = request.headers["Authorization"]
        token = self.get_token_from_header(authorization=auth, prefix=self.prefix)
        try:
            payload = jwt.decode(token, key=self.secret_key, algorithms=self.algorithm)
        except jwt.InvalidTokenError as e:
            raise AuthenticationError(str(e))

        return AuthCredentials(["authenticated"]), JWTUser(username=payload[self.username_field], token=token,
                                                           payload=payload)


class JWTWebSocketAuthenticationBackend(AuthenticationBackend):

    def __init__(self, secret_key: str, algorithm: str = 'HS256', query_param_name: str = 'jwt',
                 username_field: str = 'username'):
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.query_param_name = query_param_name
        self.username_field = username_field

    async def authenticate(self, request):
        if self.query_param_name not in request.query_params:
            return AuthCredentials(), UnauthenticatedUser()
        token = request.query_params[self.query_param_name]

        try:
            payload = jwt.decode(token, key=self.secret_key, algorithms=self.algorithm)
            print(payload)
        except jwt.InvalidTokenError as e:
            raise AuthenticationError(str(e))

        return AuthCredentials(["authenticated"]), JWTUser(username=payload[self.username_field], token=token,
                                                           payload=payload)



"""High-level Thing base class implementation."""

from jsonschema import validate
from jsonschema.exceptions import ValidationError

from websockets import ConnectionClosedOK
from starlette.websockets import WebSocketDisconnect


class Thing:
    """A Web Thing."""

    type = []
    description = ""

    def __init__(self, id_, title):
        """
        Initialize the object.
        id_ -- the thing's unique ID - must be a URI
        title -- the thing's title
        type_ -- the thing's type(s)
        description -- description of the thing
        """
        if not isinstance(self.type, list):
            self.type = [self.type]

        self.id = id_
        self.context = "https://iot.mozilla.org/schemas"
        self.title = title
        self.properties = {}
        self.available_actions = {}
        self.available_events = {}
        self.actions = {}
        self.events = []
        self.subscribers = {}
        self.href_prefix = ""
        self.ui_href = None

    async def as_thing_description(self):
        """
        Return the thing state as a Thing Description.
        Returns the state as a dictionary.
        """
        thing = {
            "id": self.id,
            "title": self.title,
            "@context": self.context,
            "properties": await self.get_property_descriptions(),
            "actions": {},
            "events": {},
            "links": [
                {"rel": "properties", "href": f"{self.href_prefix}/properties", },
                {"rel": "actions", "href": f"{self.href_prefix}/actions", },
                {"rel": "events", "href": f"{self.href_prefix}/events", },
            ],
        }

        for name, action in self.available_actions.items():
            thing["actions"][name] = action["metadata"]
            thing["actions"][name]["links"] = [
                {"rel": "action", "href": f"{self.href_prefix}/actions/{name}", },
            ]

        for name, event in self.available_events.items():
            thing["events"][name] = event["metadata"]
            thing["events"][name]["links"] = [
                {"rel": "event", "href": f"{self.href_prefix}/events/{name}", },
            ]

        if self.ui_href is not None:
            thing["links"].append(
                {"rel": "alternate", "mediaType": "text/html", "href": self.ui_href, }
            )

        if self.description:
            thing["description"] = self.description

        if self.type:
            thing["@type"] = self.type

        return thing

    async def get_href(self):
        """Get this thing's href."""
        if self.href_prefix:
            return self.href_prefix

        return "/"

    async def get_ui_href(self):
        """Get the UI href."""
        return self.ui_href

    async def set_href_prefix(self, prefix):
        """
        Set the prefix of any hrefs associated with this thing.
        prefix -- the prefix
        """
        self.href_prefix = prefix

        for property_ in self.properties.values():
            await property_.set_href_prefix(prefix)

        for action_name in self.actions.keys():
            for action in self.actions[action_name]:
                await action.set_href_prefix(prefix)

    async def set_ui_href(self, href):
        """
        Set the href of this thing's custom UI.
        href -- the href
        """
        self.ui_href = href

    async def get_id(self):
        """
        Get the ID of the thing.
        Returns the ID as a string.
        """
        return self.id

    async def get_title(self):
        """
        Get the title of the thing.
        Returns the title as a string.
        """
        return self.title

    async def get_context(self):
        """
        Get the type context of the thing.
        Returns the context as a string.
        """
        return self.context

    async def get_type(self):
        """
        Get the type(s) of the thing.
        Returns the list of types.
        """
        return self.type

    async def get_description(self):
        """
        Get the description of the thing.
        Returns the description as a string.
        """
        return self.description

    async def get_property_descriptions(self):
        """
        Get the thing's properties as a dictionary.
        Returns the properties as a dictionary, i.e. name -> description.
        """
        return {
            k: await v.as_property_description() for k, v in self.properties.items()
        }

    async def get_action_descriptions(self, action_name=None):
        """
        Get the thing's actions as an array.
        action_name -- Optional action name to get descriptions for
        Returns the action descriptions.
        """
        descriptions = []

        if action_name is None:
            print(self.actions)
            for name in self.actions:
                for action in self.actions[name]:
                    descriptions.append(await action.as_action_description())
        elif action_name in self.actions:
            for action in self.actions[action_name]:
                descriptions.append(await action.as_action_description())

        return descriptions

    async def get_event_descriptions(self, event_name=None):
        """
        Get the thing's events as an array.
        event_name -- Optional event name to get descriptions for
        Returns the event descriptions.
        """
        if event_name is None:
            return [await e.as_event_description() for e in self.events]
        else:
            return [
                await e.as_event_description()
                for e in self.events
                if await e.get_name() == event_name
            ]

    async def add_property(self, property_):
        """
        Add a property to this thing.
        property_ -- property to add
        """
        await property_.set_href_prefix(self.href_prefix)
        await property_.set_thing(self)
        self.properties[property_.name] = property_

    async def remove_property(self, property_):
        """
        Remove a property from this thing.
        property_ -- property to remove
        """
        if property_.name in self.properties:
            del self.properties[property_.name]

    async def find_property(self, property_name):
        """
        Find a property by name.
        property_name -- the property to find
        Returns a Property object, if found, else None.
        """
        return self.properties.get(property_name, None)

    async def get_property(self, property_name):
        """
        Get a property's value.
        property_name -- the property to get the value of
        Returns the properties value, if found, else None.
        """
        prop = await self.find_property(property_name)
        if prop:
            return await prop.get_value()

        return None

    async def get_properties(self):
        """
        Get a mapping of all properties and their values.
        Returns a dictionary of property_name -> value.
        """
        return {
            await prop.get_name(): await prop.get_value()
            for prop in self.properties.values()
        }

    async def has_property(self, property_name):
        """
        Determine whether or not this thing has a given property.
        property_name -- the property to look for
        Returns a boolean, indicating whether or not the thing has the
        property.
        """
        return property_name in self.properties

    async def set_property(self, property_name, value):
        """
        Set a property value.
        property_name -- name of the property to set
        value -- value to set
        """
        prop = await self.find_property(property_name)
        if not prop:
            return
        print(f"set {self.id}'s property {property_name} to {value}")
        await prop.set_value(value)

    async def sync_property(self, property_name, value):
        """
        Sync a property value from cloud or mqtt etc.
        property_name -- name of the property to set
        value -- value to set
        """
        prop = await self.find_property(property_name)
        if not prop:
            return
        print(f"sync {self.title}'s property {property_name} to {value}")
        await prop.set_value(value, with_action=False)

    async def get_action(self, action_name, action_id):
        """
        Get an action.
        action_name -- name of the action
        action_id -- ID of the action
        Returns the requested action if found, else None.
        """
        if action_name not in self.actions:
            return None

        for action in self.actions[action_name]:
            if action.id == action_id:
                return action

        return None

    async def add_event(self, event):
        """
        Add a new event and notify subscribers.
        event -- the event that occurred
        """
        self.events.append(event)
        await event.set_thing(self)
        await self.event_notify(event)

    async def add_available_event(self, cls, metadata):
        """
        Add an available event.
        name -- name of the event
        metadata -- event metadata, i.e. type, description, etc., as a dict
        """
        if metadata is None:
            metadata = {}

        self.available_events[cls.name] = {
            "metadata": metadata,
            "subscribers": {},
        }

    async def perform_action(self, action_name, input_=None):
        """
        Perform an action on the thing.
        action_name -- name of the action
        input_ -- any action inputs
        Returns the action that was created.
        """
        if action_name not in self.available_actions:
            return None

        action_type = self.available_actions[action_name]

        if "input" in action_type["metadata"]:
            try:
                validate(input_, action_type["metadata"]["input"])
            except ValidationError:
                return None

        action = action_type["class"](self, input_=input_)
        await action.set_href_prefix(self.href_prefix)
        await self.action_notify(action)
        self.actions[action_name].append(action)
        return action

    async def remove_action(self, action_name, action_id):
        """
        Remove an existing action.
        action_name -- name of the action
        action_id -- ID of the action
        Returns a boolean indicating the presence of the action.
        """
        action = await self.get_action(action_name, action_id)
        if action is None:
            return False

        await action.cancel()
        self.actions[action_name].remove(action)
        return True

    async def add_available_action(self, metadata, cls):
        """
        Add an available action.
        name -- name of the action, default use cls.name
        metadata -- action metadata, i.e. type, description, etc., as a dict
        cls -- class to instantiate for this action
        """
        if metadata is None:
            metadata = {}

        name = cls.name
        self.available_actions[name] = {
            "metadata": metadata,
            "class": cls,
        }
        self.actions[name] = []

    async def add_subscriber(self, ws):
        """
        Add a new websocket subscriber.
        ws -- the websocket
        """
        if id(ws) not in self.subscribers:
            self.subscribers[id(ws)] = ws

    async def remove_subscriber(self, ws):
        """
        Remove a websocket subscriber.
        ws -- the websocket
        """
        if id(ws) in self.subscribers:
            self.subscribers.pop(id(ws))

        for name in self.available_events:
            await self.remove_event_subscriber(name, ws)

    async def add_event_subscriber(self, name, ws):
        """
        Add a new websocket subscriber to an event.
        name -- name of the event
        ws -- the websocket
        """
        if name in self.available_events:
            if id(ws) not in self.available_events[name]["subscribers"]:
                self.available_events[name]["subscribers"][id(ws)] = ws

    async def remove_event_subscriber(self, name, ws):
        """
        Remove a websocket subscriber from an event.
        name -- name of the event
        ws -- the websocket
        """
        if (
                name in self.available_events
                and id(ws) in self.available_events[name]["subscribers"]
        ):
            self.available_events[name]["subscribers"].pop(id(ws))

    async def property_notify(self, property_):
        """
        Notify all subscribers of a property change.
        property_ -- the property that changed
        """
        message = {
            "messageType": "propertyStatus",
            "data": {property_.name: await property_.get_value(), },
        }

        for subscriber in list(self.subscribers.values()):
            try:
                await subscriber.send_json(message, mode="binary")
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass

    async def property_action(self, property_):
        """
        Addional action when a property change.
        property_ -- the property that changed
        """
        pass

    async def action_notify(self, action):
        """
        Notify all subscribers of an action status change.
        action -- the action whose status changed
        """
        message = {
            "messageType": "actionStatus",
            "data": await action.as_action_description(),
        }

        for subscriber in list(self.subscribers.values()):
            try:
                await subscriber.send_json(message, mode="binary")
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass

    async def event_notify(self, event):
        """
        Notify all subscribers of an event.
        event -- the event that occurred
        """

        if event.name not in self.available_events:
            return

        message = {
            "messageType": "event",
            "data": await event.as_event_description(),
        }

        for subscriber in list(self.available_events[event.name]["subscribers"].values()):
            try:
                await subscriber.send_json(message, mode="binary")
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass



"""A setuptools based setup module."""

from setuptools import setup, find_packages
from codecs import open
from os import path


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file.
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "ifaddr>=0.1.0",
    "pyee>=7.0.0",
    "jsonschema>=3.2.0",
    "zeroconf>=0.27.0",
    "ujson",
    "pyjwt",
    "httpx==0.13.*",
    "starlette==0.13.*",
    "uvicorn==0.11.*",
]

setup(
    name="aiowebthing",
    version="0.1.14",
    description="Web of Thing async implementation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hidaris/aiowebthing",
    author="hidaris",
    author_email="zuocool@gmail.com",
    keywords="async mozilla iot web-of-thing webthing",
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    install_requires=requirements,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    license="MPL-2.0",
    project_urls={
        "Source": "https://github.com/hidaris/aiowebthing",
        "Tracker": "https://github.com/hidaris/aiowebthing/issues",
    },
    python_requires=">=3.6, <4",
)



"""This module provides a high-level interface for creating a Web Thing."""

# flake8: noqa
from .action import Action
from .event import Event
from .property import Property
from .server import WebThingServer
from .containers import SingleThing, MultipleThings
from .thing import Thing
from .value import Value
from .utils import background_thread_loop
from .middlewares import JWTAuthenticationBackend, JWTWebSocketAuthenticationBackend
